import 'dart:typed_data';
import 'dart:math';

class AlignmentException implements Exception {
  final String msg;

  const AlignmentException([this.msg]);

  @override
  String toString() => msg ?? 'AlignmentException';
}

class MemoryError implements Error {
  final String msg;

  const MemoryError([this.msg]);
  external StackTrace get stackTrace;

  @override
  String toString() => msg ?? 'MemoryError';
}

class TVMMem {
  final int pageSZ;
  final int memSZ;
  final int pageBits;
  final int addressBits;

  List<ByteData> addressRange;

  TVMMem(memSZ, pageSZ)
    : pageSZ = pageSZ, memSZ = memSZ,
      pageBits = (log(memSZ / pageSZ) / log(2)) as int, addressBits = log(memSZ) / log(2) as int {
    addressRange = new List(memSZ / pageSZ);
    
    if(pageSZ % 4 != 0)
      throw new MemoryError("pages must be a multiple of 4 bytes");
    if(memSZ % pageSZ != 0)
      throw new MemoryError("mem size should be a multiple of page size");
    if(memSZ > 4294967296)
      throw new MemoryError("TVM only supports up to a 32bit address space");
  }

  int getPage(int address) => address >> (addressBits - pageBits);
  int getOffset(int address) => (address & (pow(2, addressBits - pageBits) - 1));
  void enablePage(int pageNum){
    if(addressRange[pageNum] == null){
      addressRange[pageNum] = new ByteData(pageSZ);
    }
  }

  void setUint8(int byteOffset, int value){
    int pageNum = getPage(byteOffset);
    int offset = getOffset(byteOffset);

    enablePage(pageNum);
    addressRange[pageNum].setUint8(offset, value);
  }

  void setUint16(int byteOffset, int value){
    int pageNum = getPage(byteOffset);
    int offset = getOffset(byteOffset);

    if(byteOffset % 2 != 0)
      throw new AlignmentException("address must be 16-bit alligned");

    enablePage(pageNum);
    addressRange[pageNum].setUint16(offset, value);
  }

  void setUint32(int byteOffset, int value){
    int pageNum = getPage(byteOffset);
    int offset = getOffset(byteOffset);

    if(byteOffset % 4 != 0)
      throw new AlignmentException("address must be 32-bit alligned");

    enablePage(pageNum);
    addressRange[pageNum].setUint32(offset, value);
  }

  int getUint8(int byteOffset, int value){
    int pageNum = getPage(byteOffset);
    int offset = getOffset(byteOffset);

    return (addressRange[pageNum] == null) ? 0 : addressRange[pageNum].getUint8(offset);
  }

  int getUint16(int byteOffset, int value){
    int pageNum = getPage(byteOffset);
    int offset = getOffset(byteOffset);

    if(byteOffset % 2 != 0)
      throw new AlignmentException("address must be 16-bit alligned");

    return (addressRange[pageNum] == null) ? 0 : addressRange[pageNum].getUint16(offset);
  }

  int getUint32(int byteOffset, int value){
    int pageNum = getPage(byteOffset);
    int offset = getOffset(byteOffset);

    if(byteOffset % 4 != 0)
      throw new AlignmentException("address must be 32-bit alligned");

    return (addressRange[pageNum] == null) ? 0 : addressRange[pageNum].getUint32(offset);
  }

  //this could probably be spead up significantly
  void loadByteData(int byteOffset, ByteData data){
    for(int i = 0; i < data.lengthInBytes; i++){
      this.setUint8(byteOffset + i, data.getUint8(i));
    }
  }  
}
